import platform
import subprocess

from absl import logging

_DARWIN = 'Darwin'
_LINUX = 'Linux'
_WINDOWS = 'Windows'


def open_browser(url: str):
  """Opens the browser to the login page."""
  system = platform.system()
  if system == _DARWIN:
    cmd = ['open', url]
  elif system == _LINUX:
    cmd = ['xdg-open', url]
  elif system == _WINDOWS:
    logging.info(
        'I can\'t imagine this is going to work on Windows, but feel free to '
        'give it a try and report back')
    cmd = ['cmd', '/c', 'start', url.replace('&', '^&')]
  else:
    raise RuntimeError(f'Unsupported system: {system}')
  try:
    subprocess.run(cmd, check=True, capture_output=True)
  except subprocess.CalledProcessError as e:
    if e.stdout:
      print(e.stdout)
    raise e
  # Unhandled: FileNotFoundError



"""Handle outputs from the server."""
import pathlib
from typing import Set

from rx.client.worker import rsync
from rx.proto import rx_pb2


class OutputHandler:
  """Handle output files and stdout/err."""

  def __init__(self, rxroot: pathlib.Path) -> None:
    self._rxroot = rxroot
    self._current_outputs: Set[pathlib.Path] = set()

  def handle(self, resp: rx_pb2.ExecResponse):
    """Creates output files."""
    for pth_str in resp.output_files:
      self._current_outputs.add(pathlib.Path(pth_str))

  def write_outputs(self, rsync_client: rsync.RsyncClient):
    if not self._current_outputs:
      return
    rsync_client.from_remote('.', self._rxroot)
    print('Changed:')
    for pth in sorted(self._current_outputs):
      print(f'  {pth}')



from typing import Iterable, Optional, cast

import grpc
import tqdm

from rx.proto import rx_pb2


def show_progress_bars(
    it: Iterable[rx_pb2.DockerImageProgress]) -> Optional[rx_pb2.Result]:
  progress_bars = {}
  try:
    for pp in it:
      if pp.id not in progress_bars:
        # First item must have a total.
        if pp.total == 0:
          continue
        progress_bars[pp.id] = ProgressBar(pp)
      progress_bars[pp.id].update(pp)
  except grpc.RpcError as e:
    e = cast(grpc.Call, e)
    return rx_pb2.Result(code=rx_pb2.UNKNOWN, message=e.details())
  finally:
    for p in progress_bars.values():
      p.close()


class ProgressBar:

  def __init__(self, pp: rx_pb2.DockerImageProgress) -> None:
    assert pp.total > 0
    self._bar = tqdm.tqdm(
      desc=f'{pp.status} layer {pp.id}',
      total=pp.total,
      unit=' bytes')
    self._is_done = False

  def close(self):
    self._bar.close()

  def update(self, pp: rx_pb2.DockerImageProgress):
    if pp.total == 0:
      self._bar.set_description(f'{pp.status} layer {pp.id}')
      return
    if self._is_done:
      return

    # Note: this jumps back to 0 for download -> extract, which is good?
    self._bar.update(pp.current - self._bar.n)
    if self._bar.n == pp.total:
      self._is_done = True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rx/proto/rx.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11rx/proto/rx.proto\x12\x02rx\x1a\x1bgoogle/protobuf/empty.proto\"\x1d\n\x08Hardware\x12\x11\n\tprocessor\x18\x01 \x01(\t\":\n\x05Image\x12\x10\n\x08registry\x18\x01 \x01(\t\x12\x12\n\nrepository\x18\x02 \x01(\t\x12\x0b\n\x03tag\x18\x03 \x01(\t\"7\n\x06Remote\x12\x1e\n\x08hardware\x18\x01 \x01(\x0b\x32\x0c.rx.Hardware\x12\r\n\x05ports\x18\x02 \x03(\x05\"\"\n\x12\x45nvironmentManager\x12\x0c\n\x04name\x18\x01 \x01(\t\"q\n\x0b\x45nvironment\x12\x1a\n\x06remote\x18\x01 \x01(\x0b\x32\n.rx.Remote\x12\x18\n\x05image\x18\x02 \x01(\x0b\x32\t.rx.Image\x12,\n\x0c\x65nv_managers\x18\x03 \x03(\x0b\x32\x16.rx.EnvironmentManager\"?\n\x05\x44\x65lta\x12\x10\n\x08\x61\x64\x64_path\x18\x01 \x03(\t\x12\x0f\n\x07\x61\x64\x64_dir\x18\x02 \x03(\t\x12\x13\n\x0bremove_path\x18\x03 \x03(\t\"Q\n\x13\x44ockerImageProgress\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\r\n\x05total\x18\x03 \x01(\x03\x12\x0f\n\x07\x63urrent\x18\x04 \x01(\x03\"%\n\rSubscribeInfo\x12\x14\n\x0cpayment_link\x18\x01 \x01(\t\"7\n\x06Result\x12\x1c\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x0e.rx.StatusCode\x12\x0f\n\x07message\x18\x02 \x01(\t\"H\n\x0bRsyncSource\x12\x12\n\nmachine_id\x18\x01 \x01(\x03\x12\x11\n\tdirectory\x18\x02 \x01(\t\x12\x12\n\npublic_key\x18\x03 \x01(\x0c\")\n\x10RsyncDestination\x12\x15\n\rdaemon_module\x18\x01 \x01(\t\":\n\tExecution\x12\x0b\n\x03\x63md\x18\x01 \x01(\t\x12\x10\n\x08start_ts\x18\x02 \x01(\x03\x12\x0e\n\x06\x65nd_ts\x18\x03 \x01(\x03\"-\n\x0fGenericResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\"x\n\x0b\x45xecRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgv\x18\x02 \x03(\t\x12)\n\x0crsync_source\x18\x03 \x01(\x0b\x32\x0f.rx.RsyncSourceB\x02\x18\x01\x12\x0b\n\x03\x63wd\x18\x04 \x01(\t\x12\r\n\x05stdin\x18\x05 \x01(\x0c\"\x89\x01\n\x0c\x45xecResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x14\n\x0c\x65xecution_id\x18\x02 \x01(\t\x12\x0e\n\x06stdout\x18\x03 \x01(\x0c\x12\x0e\n\x06stderr\x18\x04 \x01(\x0c\x12\x14\n\x0coutput_files\x18\x05 \x03(\t\x12\x11\n\texit_code\x18\x06 \x01(\x05\"a\n\x18GetSubscribeInfoResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12)\n\x0esubscribe_info\x18\x02 \x01(\x0b\x32\x11.rx.SubscribeInfo\"?\n\x0fGetUserResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x10\n\x08username\x18\x02 \x01(\t\"/\n\x17GetWorkspaceInfoRequest\x12\x14\n\x0cworkspace_id\x18\x02 \x01(\t\"\x8b\x01\n\x18GetWorkspaceInfoResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\r\n\x05state\x18\x02 \x01(\t\x12\x1e\n\x07history\x18\x03 \x03(\x0b\x32\r.rx.Execution\x12$\n\x0b\x65nvironment\x18\x04 \x01(\x0b\x32\x0f.rx.Environment\"o\n\x0bInitRequest\x12\x14\n\x0cproject_name\x18\x01 \x01(\t\x12%\n\x0crsync_source\x18\x02 \x01(\x0b\x32\x0f.rx.RsyncSource\x12#\n\ntarget_env\x18\x03 \x01(\x0b\x32\x0f.rx.Environment\"\x7f\n\x0cInitResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12(\n\nrsync_dest\x18\x02 \x01(\x0b\x32\x14.rx.RsyncDestination\x12\x13\n\x0bworker_addr\x18\x03 \x01(\t\x12\x14\n\x0cworkspace_id\x18\x04 \x01(\t\"*\n\x12InstallDepsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"A\n\x13InstallDepsResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x0e\n\x06stdout\x18\x02 \x01(\x0c\"9\n\x0bKillRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x14\n\x0c\x65xecution_id\x18\x02 \x01(\t\"\'\n\x0fUnfreezeRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"C\n\x10UnfreezeResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12\x13\n\x0bworker_addr\x18\x02 \x01(\t\"&\n\x12SetUsernameRequest\x12\x10\n\x08username\x18\x01 \x01(\t\"1\n\x0bStopRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x0c\n\x04save\x18\x02 \x01(\x08\"t\n\x0cStopResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12.\n\rpush_progress\x18\x02 \x01(\x0b\x32\x17.rx.DockerImageProgress\x12\x18\n\x05image\x18\x03 \x01(\x0b\x32\t.rx.Image\")\n\x11WorkerInitRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"`\n\x12WorkerInitResponse\x12\x1a\n\x06result\x18\x01 \x01(\x0b\x32\n.rx.Result\x12.\n\rpull_progress\x18\x02 \x01(\x0b\x32\x17.rx.DockerImageProgress*\x82\x01\n\nStatusCode\x12\x06\n\x02OK\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x01\x12\x0b\n\x07INVALID\x10\x02\x12\r\n\tNOT_FOUND\x10\x03\x12\x10\n\x0cUNAUTHORIZED\x10\x04\x12\x19\n\x15SUBSCRIPTION_REQUIRED\x10\x05\x12\n\n\x06\x45\x41GAIN\x10\x0b\x12\n\n\x05MOVED\x10\xad\x02\x32\xf5\x01\n\x10\x45xecutionService\x12\x39\n\x04Init\x12\x15.rx.WorkerInitRequest\x1a\x16.rx.WorkerInitResponse\"\x00\x30\x01\x12\x42\n\x0bInstallDeps\x12\x16.rx.InstallDepsRequest\x1a\x17.rx.InstallDepsResponse\"\x00\x30\x01\x12/\n\x04\x45xec\x12\x0f.rx.ExecRequest\x1a\x10.rx.ExecResponse\"\x00(\x01\x30\x01\x12\x31\n\x04Kill\x12\x0f.rx.KillRequest\x1a\x16.google.protobuf.Empty\"\x00\x32\xbc\x04\n\x0cSetupService\x12\x42\n\x11\x43heckSubscription\x12\x16.google.protobuf.Empty\x1a\x13.rx.GenericResponse\"\x00\x12J\n\x10GetSubscribeInfo\x12\x16.google.protobuf.Empty\x1a\x1c.rx.GetSubscribeInfoResponse\"\x00\x12\x38\n\x07GetUser\x12\x16.google.protobuf.Empty\x1a\x13.rx.GetUserResponse\"\x00\x12O\n\x10GetWorkspaceInfo\x12\x1b.rx.GetWorkspaceInfoRequest\x1a\x1c.rx.GetWorkspaceInfoResponse\"\x00\x12+\n\x04Init\x12\x0f.rx.InitRequest\x1a\x10.rx.InitResponse\"\x00\x12\x39\n\x08Unfreeze\x12\x13.rx.UnfreezeRequest\x1a\x14.rx.UnfreezeResponse\"\x00\x30\x01\x12<\n\x0bSetUsername\x12\x16.rx.SetUsernameRequest\x1a\x13.rx.GenericResponse\"\x00\x12-\n\x04Stop\x12\x0f.rx.StopRequest\x1a\x10.rx.StopResponse\"\x00\x30\x01\x12<\n\x0bUnsubscribe\x12\x16.google.protobuf.Empty\x1a\x13.rx.GenericResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rx.proto.rx_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EXECREQUEST.fields_by_name['rsync_source']._options = None
  _EXECREQUEST.fields_by_name['rsync_source']._serialized_options = b'\030\001'
  _STATUSCODE._serialized_start=2311
  _STATUSCODE._serialized_end=2441
  _HARDWARE._serialized_start=54
  _HARDWARE._serialized_end=83
  _IMAGE._serialized_start=85
  _IMAGE._serialized_end=143
  _REMOTE._serialized_start=145
  _REMOTE._serialized_end=200
  _ENVIRONMENTMANAGER._serialized_start=202
  _ENVIRONMENTMANAGER._serialized_end=236
  _ENVIRONMENT._serialized_start=238
  _ENVIRONMENT._serialized_end=351
  _DELTA._serialized_start=353
  _DELTA._serialized_end=416
  _DOCKERIMAGEPROGRESS._serialized_start=418
  _DOCKERIMAGEPROGRESS._serialized_end=499
  _SUBSCRIBEINFO._serialized_start=501
  _SUBSCRIBEINFO._serialized_end=538
  _RESULT._serialized_start=540
  _RESULT._serialized_end=595
  _RSYNCSOURCE._serialized_start=597
  _RSYNCSOURCE._serialized_end=669
  _RSYNCDESTINATION._serialized_start=671
  _RSYNCDESTINATION._serialized_end=712
  _EXECUTION._serialized_start=714
  _EXECUTION._serialized_end=772
  _GENERICRESPONSE._serialized_start=774
  _GENERICRESPONSE._serialized_end=819
  _EXECREQUEST._serialized_start=821
  _EXECREQUEST._serialized_end=941
  _EXECRESPONSE._serialized_start=944
  _EXECRESPONSE._serialized_end=1081
  _GETSUBSCRIBEINFORESPONSE._serialized_start=1083
  _GETSUBSCRIBEINFORESPONSE._serialized_end=1180
  _GETUSERRESPONSE._serialized_start=1182
  _GETUSERRESPONSE._serialized_end=1245
  _GETWORKSPACEINFOREQUEST._serialized_start=1247
  _GETWORKSPACEINFOREQUEST._serialized_end=1294
  _GETWORKSPACEINFORESPONSE._serialized_start=1297
  _GETWORKSPACEINFORESPONSE._serialized_end=1436
  _INITREQUEST._serialized_start=1438
  _INITREQUEST._serialized_end=1549
  _INITRESPONSE._serialized_start=1551
  _INITRESPONSE._serialized_end=1678
  _INSTALLDEPSREQUEST._serialized_start=1680
  _INSTALLDEPSREQUEST._serialized_end=1722
  _INSTALLDEPSRESPONSE._serialized_start=1724
  _INSTALLDEPSRESPONSE._serialized_end=1789
  _KILLREQUEST._serialized_start=1791
  _KILLREQUEST._serialized_end=1848
  _UNFREEZEREQUEST._serialized_start=1850
  _UNFREEZEREQUEST._serialized_end=1889
  _UNFREEZERESPONSE._serialized_start=1891
  _UNFREEZERESPONSE._serialized_end=1958
  _SETUSERNAMEREQUEST._serialized_start=1960
  _SETUSERNAMEREQUEST._serialized_end=1998
  _STOPREQUEST._serialized_start=2000
  _STOPREQUEST._serialized_end=2049
  _STOPRESPONSE._serialized_start=2051
  _STOPRESPONSE._serialized_end=2167
  _WORKERINITREQUEST._serialized_start=2169
  _WORKERINITREQUEST._serialized_end=2210
  _WORKERINITRESPONSE._serialized_start=2212
  _WORKERINITRESPONSE._serialized_end=2308
  _EXECUTIONSERVICE._serialized_start=2444
  _EXECUTIONSERVICE._serialized_end=2689
  _SETUPSERVICE._serialized_start=2692
  _SETUPSERVICE._serialized_end=3264
# @@protoc_insertion_point(module_scope)



import pathlib

from rx.client.configuration import config_base
from rx.client.configuration import local
from rx.client.configuration import remote

class Command:
  def __init__(self) -> None:
    if config_base.RX_ROOT.value:
      self._rxroot = pathlib.Path(config_base.RX_ROOT.value)
    else:
      self._rxroot = local.find_rxroot(pathlib.Path.cwd())

  @property
  def local_config(self) -> local.LocalConfig:
    if not self._rxroot:
      raise config_base.ConfigNotFoundError(
        pathlib.Path('.'), 'Run `rx init` first!')
    config = local.find_local_config(self._rxroot)
    if not config:
      raise config_base.ConfigNotFoundError(
        self._rxroot, 'Run `rx init` first!')
    return config

  @property
  def remote_config(self) -> remote.Remote:
    return remote.Remote(self.local_config.cwd)



import argparse
from typing import List

from absl import logging

from rx.client import trex_client
from rx.client import worker_client
from rx.client import grpc_helper
from rx.client.commands import command
from rx.client.configuration import config_base


class RunCommand(command.Command):

  def __init__(self, argv: List[str]):
    super().__init__()
    self._argv = argv

  def run(self) -> int:
    try:
      with grpc_helper.get_channel(self.remote_config.worker_addr) as ch:
        client = worker_client.create_authed_client(ch, self.local_config)
        return self._try_exec(client)
    except config_base.ConfigNotFoundError as e:
      print(e)
      return -1

  def _try_exec(self, client: worker_client.Client) -> int:
    """Sends the command to the server."""
    if len(self._argv) < 1:
      print('No command given.')
      return 1

    if len(self._argv) == 1:
      # If this is a quoted arg, break it into pieces. E.g., to pass "ls -l" it
      # must be quoted so that absl doesn't try to capture the -l.
      # TODO: is there a better way to handle this with absl flags?
      self._argv = self._argv[0].split(' ')
    try:
      return client.exec(list(self._argv))
    except KeyboardInterrupt:
      client.maybe_kill()
      return worker_client.SIGINT_CODE
    except worker_client.RsyncError as e:
      # Fallthrough - retry
      logging.info('Rsync error: %s', e)
      pass
    except worker_client.WorkerError as e:
      print(e)
      return e.code

    # Fallthrough from retry.
    print('Error syncing code to your worker, checking with the scheduler...')
    return self.maybe_unfreeze()


  def maybe_unfreeze(self) -> int:
    # Connect to trex.
    with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
      cli = trex_client.create_authed_client(ch, self.local_config)
      try:
        result = cli.get_info(self.remote_config.workspace_id)
      except trex_client.TrexError as e:
        print(f'{e}, run `rx init` to get a new instance')
        return -1
      if result.state == 'frozen':
        result = cli.unfreeze(self.remote_config.workspace_id)
      else:
        print('Worker was unrechable, run `rx init` to get a new instance.')
        return -1
    if result.code == 0:
      print(
        'Done! Please rerun this command to use your newly restored workspace.')
    else:
      print(
        f'Error setting up this workspace on a new machine: {result.message}')
    return result.code


def _run_cmd(args: List[str]) -> int:
  cmd = RunCommand(args)
  return cmd.run()


def add_parser(subparsers: argparse._SubParsersAction):
  (
    subparsers
    .add_parser('run', help='Runs a command')
    .set_defaults(func=_run_cmd)
  )



import errno
import fcntl
import os
import sys
import selectors
import termios
import threading
import tty
from types import TracebackType
from typing import Any, BinaryIO, Optional, Tuple, Type

from absl import logging

from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc
from rx.client import output_handler


class Executor:

  def __init__(
      self,
      stub: rx_pb2_grpc.ExecutionServiceStub,
      req: rx_pb2.ExecRequest) -> None:
    self._stub = stub
    self._request = req

  def run(
      self,
      metadata: Tuple[Tuple[str, Any], ...],
      out_handler: output_handler.OutputHandler,
  ) -> rx_pb2.ExecResponse:
    response = None
    with StdinIterator(self._request) as req_it:
      for response in self._stub.Exec(req_it, metadata=metadata):
        self.write(response.stdout, sys.stdout.buffer)
        self.write(response.stderr, sys.stderr.buffer)
        out_handler.handle(response)
    assert response
    return response

  def write(self, buf: bytes, sink: BinaryIO) -> None:
    if not buf:
      return
    written = 0
    while written < len(buf):
      try:
        written += sink.write(buf[written:])
        sink.flush()
      except BlockingIOError as e:
        if e.errno == errno.EAGAIN:
          continue
        logging.info(
          'Error writing at byte %s of %s to %s: %s', written, len(buf),
          sink.name, e)
        raise e


class StdinIterator:

  def __init__(self, request: rx_pb2.ExecRequest) -> None:
    self._running = True
    self._stdin_buffer = [request]
    self._stdin_avail = threading.Condition()
    self._original_flags = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
    self._original_attrs = termios.tcgetattr(sys.stdin)
    self._selector = selectors.DefaultSelector()

  def __enter__(self) -> 'StdinIterator':
    # This also sets O_NONBLOCK for stdout/stderr, as they are copied from
    # stdin.
    fcntl.fcntl(sys.stdin, fcntl.F_SETFL, self._original_flags | os.O_NONBLOCK)
    # Prevent echo of characters. This is unset by the termios.tcsetattr call in
    # __exit__.
    tty.setcbreak(sys.stdin)
    self._selector.register(sys.stdin, selectors.EVENT_READ, self.got_stdin)
    th = threading.Thread(target=self.loop, daemon=True)
    th.start()
    return self

  def __exit__(self, exctype: Optional[Type[BaseException]],
             excinst: Optional[BaseException],
             exctb: Optional[TracebackType]):
    self._selector.unregister(sys.stdin)
    termios.tcsetattr(sys.stdin, termios.TCSANOW, self._original_attrs)
    fcntl.fcntl(sys.stdin, fcntl.F_SETFL, self._original_flags)
    self._running = False
    return False

  def __next__(self) -> rx_pb2.ExecRequest:
    # This is probably unnecessary, nothing should call next once exec is done.
    if not self._running:
      raise StopIteration()
    with self._stdin_avail:
      while not self._stdin_buffer:
        if not self._running:
          raise StopIteration()
        self._stdin_avail.wait()
      return self._stdin_buffer.pop(0)

  def loop(self):

    while self._running:
      events = self._selector.select(timeout=0)
      for key, mask in events:
        del mask
        key.data()

  def got_stdin(self):
    stdin_str = sys.stdin.read()
    if stdin_str:
      with self._stdin_avail:
        self._stdin_buffer += [
          rx_pb2.ExecRequest(stdin=stdin_str.encode('utf-8'))]
        self._stdin_avail.notify()



import errno
import os
import pathlib
import sys
import threading
import time
from typing import Callable, Generator, Iterable, List, Optional, Tuple, TypeVar, cast

from absl import flags
from absl import logging
import grpc

from rx.client import login
from rx.client import output_handler
from rx.client.configuration import local
from rx.client.configuration import remote
from rx.client.worker import executor
from rx.client.worker import progress_bar
from rx.client.worker import rsync
from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

SIGINT_CODE = 130

# TODO: set up UI for looking at previous executions.
_BASE_URI = 'https://la-brea.run-rx.com'
_NOT_REACHABLE = 1

_CWD = flags.DEFINE_string('cwd', None, 'Directory to run the command from')

class Client:
  """Handle contacting the remote server."""

  def __init__(
      self,
      channel: grpc.Channel,
      local_cfg: local.LocalConfig,
      auth_metadata: Tuple[Tuple[str, str]]):
    self._local_cfg = local_cfg
    self._remote_cfg = remote.Remote(local_cfg.cwd)
    self._rsync = rsync.RsyncClient(local_cfg, self._remote_cfg)
    self._stub = rx_pb2_grpc.ExecutionServiceStub(channel)
    self._metadata = local.get_grpc_metadata() + auth_metadata
    self._current_execution_id = None

  def init(self):
    # Get the container downloaded/running.
    req = rx_pb2.WorkerInitRequest(workspace_id=self._remote_cfg.workspace_id)
    resp = self._stub.Init(req, metadata=self._metadata)
    def get_progress(
        resp: Iterable[rx_pb2.WorkerInitResponse]
    ) -> Generator[rx_pb2.DockerImageProgress, None, None]:
      for r in resp:
        if r.result.code != rx_pb2.OK:
          raise WorkerError(result=r.result)
        if r.pull_progress:
          yield r.pull_progress
    result = progress_bar.show_progress_bars(get_progress(resp))
    if result and result.code != rx_pb2.OK:
      if result.code == rx_pb2.MOVED:
        raise WorkspaceRelocationError()
      raise WorkerError(
        f'Error initializing worker {self._remote_cfg.worker_addr}', result)

    # Sync sources.
    prog = ShowLongRunningProgress(title='Syncing directory contents')
    result = prog.run(self._rsync.to_remote)
    if result != 0:
      logging.info('error: %s', errno.errorcode[result])
      raise RsyncError()

    # Install deps.
    self._install_deps()

  def exec(self, argv: List[str]) -> int:
    cmd_str = ' '.join(argv)
    logging.info(f'Running `{cmd_str}` on {self._remote_cfg.worker_addr}')

    result = self._rsync.to_remote()
    if result != 0:
      raise RsyncError()

    rxroot = os.path.abspath(self._local_cfg.cwd)
    cwd = os.path.abspath(pathlib.Path.cwd())
    if is_subdir(parent=rxroot, child=cwd):
      cwd = str(pathlib.Path.cwd().relative_to(self._local_cfg.cwd))
    else:
      cwd = _CWD.value

    out_handler = output_handler.OutputHandler(self._local_cfg.cwd)

    request = rx_pb2.ExecRequest(
      workspace_id=self._remote_cfg.workspace_id, argv=argv, cwd=cwd)
    runner = executor.Executor(self._stub, request)
    try:
      response = runner.run(self._metadata, out_handler)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      sys.stderr.write(f'Error contacting {self._remote_cfg.worker_addr}: {e.details()}\n')
      return _NOT_REACHABLE
    except KeyboardInterrupt:
      self.maybe_kill()
      return SIGINT_CODE
    except RuntimeError as e:
      # Bug in grpc, probably.
      msg = str(e)
      if msg == 'cannot release un-acquired lock':
        self.maybe_kill()
        return SIGINT_CODE
      else:
        sys.stderr.write(msg)
        return -1

    if response.result.code == rx_pb2.SUBSCRIPTION_REQUIRED:
      print("""
You need a subscription to continue to use compute!

Please run `rx subscribe` to continue.""")
      return 0
    if response.result.code in [rx_pb2.MOVED, rx_pb2.EAGAIN]:
      raise WorkspaceRelocationError()
    if response.result.code != rx_pb2.OK:
      sys.stderr.write(f'{response.result.message}\n')
      return response.result.code

    out_handler.write_outputs(self._rsync)

    # Return the process's exit code.
    return response.exit_code

  def maybe_kill(self):
    """Kill the process, if it exists."""
    if self._current_execution_id is None:
      logging.error('No ID to kill.')
      return

    req = rx_pb2.KillRequest(
      workspace_id=self._remote_cfg.workspace_id,
      execution_id=self._current_execution_id,
    )
    self._stub.Kill(req, metadata=self._metadata)

  def _install_deps(self):
    req = rx_pb2.InstallDepsRequest(workspace_id=self._remote_cfg.workspace_id)
    response = None
    try:
      for response in self._stub.InstallDeps(
        req, metadata=self._metadata, timeout=(10 * 60)):
        if response.stdout:
          sys.stdout.buffer.write(response.stdout)
          sys.stdout.buffer.flush()
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise WorkerError(e.details(), result=None)
    assert response, 'InstallDeps should always return something'
    if response.result.code == rx_pb2.MOVED:
      raise WorkspaceRelocationError()
    if response.result.code:
      raise WorkerError(message=None, result=response.result)


def create_authed_client(ch: grpc.Channel, local_cfg: local.LocalConfig):
  lm = login.LoginManager(local_cfg.cwd)
  lm.login()
  return Client(ch, local_cfg, lm.grpc_metadata)


def is_subdir(*, parent: str, child: str) -> bool:
  return os.path.commonpath([parent]) == os.path.commonpath([parent, child])


class WorkerError(RuntimeError):
  def __init__(
      self,
      message: Optional[str] = None,
      result: Optional[rx_pb2.Result] = None):
    """Allow passing message, result, or both."""
    full_message = message
    if message is not None and result is not None:
      full_message = f'{message}: {result.message}'
    elif result is not None:
      full_message = result.message
    super().__init__(full_message)
    self.code = result.code if result is not None else -1

T = TypeVar('T')

# TODO: use semaphores.
class ShowLongRunningProgress:
  def __init__(
      self, title: str, secs: int = 1, message: str = '.'
  ):
    self._title = title
    self._sleep_secs = secs
    self._message = message
    self._still_running = True

  def run(self, func: Callable[[], T]) -> T:
    th = threading.Thread(target=self._message_printer, daemon=True)
    th.start()
    retval = func()
    self._still_running = False
    return retval

  def _message_printer(self):
    first = True
    time.sleep(self._sleep_secs)
    while self._still_running:
      if first:
        sys.stdout.write(self._title)
        first = False
      sys.stdout.write(self._message)
      sys.stdout.flush()
      time.sleep(self._sleep_secs)


class RsyncError(WorkerError):
  def __init__(self, *args: object):
    super().__init__('unreachable', None, *args)


class WorkspaceRelocationError(RuntimeError):
  """The worker was moved to another host."""
  pass



import argparse
import sys
from typing import List

import yaml

from rx.client import grpc_helper
from rx.client import trex_client
from rx.client.commands import command
from rx.client.configuration import config_base

class StopCommand(command.Command):
  """Initialize (or reinitialize) the remote."""

  def run(self) -> int:
    try:
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        sys.stdout.write('Stopping your workspace...\n')
        client = trex_client.create_authed_client(ch, self.local_config)
        image = client.stop(self.remote_config.workspace_id)
    except config_base.ConfigNotFoundError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return -1
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code
    sys.stdout.write('Your remote machine has been shut down.\n')
    if image:
      print(
        '\nIf you\'d like to initialize a new workspace with this state, '
        f'use the following lines in your config:\n\n{yaml.safe_dump(image)}')
    return 0


def _run_cmd(args: List[str]) -> int:
  del args
  cmd = StopCommand()
  return cmd.run()


def add_parser(subparsers: argparse._SubParsersAction):
  (
    subparsers
    .add_parser('stop', help='Stops the current instance')
    .set_defaults(func=_run_cmd)
  )


if __name__ == '__main__':
  print('Call exec.')



import datetime
import sys
import time
from typing import cast, Any, Dict, Generator, Iterable, Optional, Tuple

from absl import logging
from google.protobuf import empty_pb2
from google.protobuf import json_format
import grpc

from rx.client import grpc_helper
from rx.client import login
from rx.client import payment
from rx.client import user
from rx.client import worker_client
from rx.client.configuration import local
from rx.client.configuration import remote
from rx.client.worker import progress_bar
from rx.client.worker import rsync
from rx.proto import rx_pb2
from rx.proto import rx_pb2_grpc

_PAYMENT_TIMEOUT = datetime.timedelta(300)
_RPC_TIMEOUT_SECS = seconds=10
_SLEEP_SECS = 5

class Client:
  """Handle contacting the remote server."""

  def __init__(
      self,
      channel: grpc.Channel,
      local_cfg: local.LocalConfig,
      auth_metadata: Optional[Tuple[Tuple[str, str]]]):
    self._stub = rx_pb2_grpc.SetupServiceStub(channel)
    self._local_cfg = local_cfg
    self._metadata = local.get_grpc_metadata()
    if auth_metadata:
      self._metadata += auth_metadata

  def create_user_or_log_in(self) -> user.User:
    """This logs in the user, then checks if trex has a username for them."""

    # First, make sure we're logged in with Google.
    lm = login.LoginManager(self._local_cfg.cwd)
    try:
      lm.login()
    except login.AuthError as e:
      raise TrexError(e, -1)
    self._metadata += lm.grpc_metadata
    email = lm.id_token['email']

    # Check if user is set up.
    # TODO: it would be better to read the username from the config and
    # automatically try to set it on the remote. However, this generally
    # shouldn't come up for normal users.
    username = self._get_username(email)
    if not user.has_config(self._local_cfg.cwd):
      with user.CreateUser(self._local_cfg.cwd) as c:
        c['username'] = username
        c['email'] = email
    return user.User(self._local_cfg.cwd, email)

  def dry_run(self):
    rsync.dry_run(self._local_cfg)

  def get_info(self, workspace_id: str) -> rx_pb2.GetWorkspaceInfoResponse:
    request = rx_pb2.GetWorkspaceInfoRequest(workspace_id=workspace_id)
    try:
      response = self._stub.GetWorkspaceInfo(request, metadata=self._metadata)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(
        f'Error reactivating workspace: {e.details()} Try running rx init '
        'again', -1)
    if response.result.code != rx_pb2.OK:
      raise TrexError(response.result.message, response.result.code)
    return response

  def init(self) -> int:
    try:
      target_env = self._local_cfg.get_target_env()
    except local.ConfigError as e:
      raise TrexError(str(e), -1)
    req = rx_pb2.InitRequest(
      project_name=self._local_cfg['project_name'],
      rsync_source=self._local_cfg.rsync_source,
      target_env=target_env,
    )
    # TODO: create a threaded UserStatus class with __enter__/__exit__.
    sys.stdout.write('Finding a remote worker... ')
    sys.stdout.flush()
    try:
      # Five minute timeout.
      resp = self._stub.Init(req, metadata=self._metadata, timeout=(5 * 60))
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not initialize worker: {e.details()}', -1)
    if resp.result.code != 0:
      if resp.result.code == rx_pb2.SUBSCRIPTION_REQUIRED:
        print("""

Whoops, you\'ll need a subscription to continue! Please run:

    $ rx subscribe

Then retry this command.
""")
        return -1
      else:
        raise TrexError(resp.result.message, -1)
    sys.stdout.write('Done.\n')

    with remote.WritableRemote(self._local_cfg.cwd) as r:
      r['workspace_id'] = resp.workspace_id
      r['worker_addr'] = resp.worker_addr
      r['daemon_module'] = resp.rsync_dest.daemon_module

    # Create a container on the worker.
    print('Setting up the container...')
    logging.info('Connecting to worker %s', resp.worker_addr)
    try:
      with grpc_helper.get_channel(resp.worker_addr) as ch:
        worker = worker_client.create_authed_client(ch, self._local_cfg)
        worker.init()
      print('\nDone setting up rx! To use, run:\n\n\t$ rx <your command>\n')
      return 0
    except worker_client.WorkerError as e:
      raise TrexError(f'Error setting up worker {resp.worker_addr}: {e}', -1)

  def subscribe(self) -> bool:
    if not payment.explain_subscription(self._local_cfg.cwd):
      return False
    # Get the link for payment.
    req = empty_pb2.Empty()
    response = self._stub.GetSubscribeInfo(req, metadata=self._metadata)
    payment.open_browser(response.subscribe_info)
    self._wait_for_payment()
    return True

  def unfreeze(self, workspace_id: str) -> rx_pb2.Result:
    print(
      'Your workspace was stowed, please stand by while it is set up on a new '
      'machine.')
    request = rx_pb2.UnfreezeRequest(workspace_id=workspace_id)
    status = None
    for status in self._stub.Unfreeze(request, metadata=self._metadata):
      sys.stdout.write('.')
    sys.stdout.write('\n')
    assert status
    if status.result.code == rx_pb2.OK:
      with remote.WritableRemote(self._local_cfg.cwd) as r:
        r['worker_addr'] = status.worker_addr
    return status.result

  def stop(self, workspace_id: str) -> Dict[str, Any]:
    logging.info('Stopping workspace %s', workspace_id)
    req = rx_pb2.StopRequest(workspace_id=workspace_id, save=True)
    x = {'image': {}}
    def get_progress(
        resp: Iterable[rx_pb2.StopResponse]
    ) -> Generator[rx_pb2.DockerImageProgress, None, None]:
      for r in resp:
        if r.result.code != 0:
          raise TrexError(r.result.message, r.result.code)
        if r.HasField('push_progress'):
          yield r.push_progress
        if r.HasField('image'):
          x['image'] = json_format.MessageToDict(r.image)
    resp = self._stub.Stop(req, metadata=self._metadata)
    result = progress_bar.show_progress_bars(get_progress(resp))
    if result and result.code != 0:
      raise TrexError('Error stopping worker', result)
    return x

  def unsubscribe(self):
    req = empty_pb2.Empty()
    response = self._stub.Unsubscribe(req, metadata=self._metadata)
    if response.result.code != rx_pb2.OK:
      raise TrexError(response.result.message, response.result.code)

  def _wait_for_payment(self):
    end = datetime.datetime.now() + _PAYMENT_TIMEOUT
    sys.stdout.write('\nWaiting for subscription to be activated...')
    while datetime.datetime.now() < end:
      sys.stdout.write('.')
      sys.stdout.flush()
      try:
        response = self._stub.CheckSubscription(
          empty_pb2.Empty(), metadata=self._metadata)
      except grpc.RpcError as e:
        e = cast(grpc.Call, e)
        raise TrexError(
          f'Could not check subscription status: {e.details()}', -1)
      if response.result.code == rx_pb2.OK:
        payment.success()
        return
      if response.result.code != rx_pb2.SUBSCRIPTION_REQUIRED:
        # Something else went wrong.
        raise TrexError(response.result.message, response.result.code)
      # Otherwise, we're just waiting for them to finish subscribing.
      time.sleep(_SLEEP_SECS)
    sys.stdout.write('\n')
    raise TimeoutError(
      'Timed out waiting for subscription to activate. Please run the command '
      'again after subscribing.')

  def _create_username(self, email: str) -> str:
    username = user.username_prompt(email)
    req = rx_pb2.SetUsernameRequest(username=username)
    try:
      resp = self._stub.SetUsername(
        req, metadata=self._metadata, timeout=_RPC_TIMEOUT_SECS)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not set username: {e.details()}', -1)
    if resp.result.code == rx_pb2.INVALID:
      raise TrexError(
        f'{resp.result.message}\nInvalid username: {username}', rx_pb2.INVALID)
    return username

  def _get_username(self, email: str) -> str:
    # Check with rx server.
    username = self._get_username_from_rx()
    if username:
      return username

    # Prompt the user to choose a username.
    return self._create_username(email)

  def _get_username_from_rx(self) -> str:
    try:
      resp = self._stub.GetUser(
        empty_pb2.Empty(), metadata=self._metadata, timeout=_RPC_TIMEOUT_SECS)
    except grpc.RpcError as e:
      e = cast(grpc.Call, e)
      raise TrexError(f'Could not get user from rx: {e.details()}', -1)
    return resp.username

  def _run_initial_rsync(self, remote_cfg: remote.Remote):
    r = rsync.RsyncClient(self._local_cfg, remote_cfg)
    return_code = r.to_remote()
    if return_code == 0:
      logging.info('Copied files to %s', r.host)


class TrexError(RuntimeError):
  """Class to repackage any init errors that happen and add an exit code."""

  def __init__(self, message, code, *args):
    super().__init__(message, *args)
    self._code = code

  @property
  def code(self):
    return self._code


def create_authed_client(ch: grpc.Channel, local_cfg: local.LocalConfig):
  lm = login.LoginManager(local_cfg.cwd)
  lm.login()
  return Client(ch, local_cfg, lm.grpc_metadata)



import argparse
import sys
from typing import List

from rx.client import grpc_helper
from rx.client import trex_client
from rx.client.commands import command
from rx.client.configuration import config_base


class SubscribeCommand(command.Command):
  """Create a subscription."""

  def run(self) -> int:
    try:
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = trex_client.create_authed_client(ch, self.local_config)
        client.subscribe()
    except config_base.ConfigNotFoundError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return -1
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code
    return 0


class UnsubscribeCommand(command.Command):
  """Cancel a subscription."""

  def run(self) -> int:
    try:
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = trex_client.create_authed_client(ch, self.local_config)
        client.unsubscribe()
    except config_base.ConfigNotFoundError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return -1
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code
    return 0


def _sub(args: List[str]) -> int:
  del args
  s = SubscribeCommand()
  return s.run()


def _unsub(args: List[str]) -> int:
  del args
  s = UnsubscribeCommand()
  return s.run()


def add_parsers(subparsers: argparse._SubParsersAction):
  (
    subparsers
    .add_parser(
      'subscribe', help='Creates up a new subscription to access rx resources')
    .set_defaults(func=_sub)
  )
  (
    subparsers
    .add_parser('unsubscribe', help='Cancels an existing subscription')
    .set_defaults(func=_unsub)
  )



if __name__ == '__main__':
  print('Call exec.')



import pathlib

from rx.client import browser
from rx.client import menu
from rx.client import user
from rx.proto import rx_pb2


_GPU = """
Subscribing includes:

    * $19.99/month for unlimited CPU usage and workspace storage.
    * $3/hour for using an H100 GPU.

GPU usage is only charged for hours when you have `rx run` commands on a GPU
workspace. rx will automatically stop charging you if you aren't using a GPU.
You can run `rx workspace-info` to check if your workspace is using a GPU
(without incurring any charges).

Finally, canceling your subscription is hassle-free: run `rx unsubscribe` at any
time.

See https://run-rx.com/pricing for more info.
"""


def explain_subscription(cwd: pathlib.Path) -> bool:
  """Returns if the user subscribed."""
  if not user.has_config(cwd):
    raise RuntimeError('No user info found, try running rx init first.')

  pay = menu.bool_prompt(f"""

Create a subscription
=====================
{_GPU}
Would you like to continue? (This will open a browser window to collect payment
infomation.)""", 'y')

  if not pay:
    print("""
Sorry to hear that! Please email us at at eng@run-rx.com know if you have any
feedback or suggestions.""")
    return False
  return True


def open_browser(sub: rx_pb2.SubscribeInfo):
  url = sub.payment_link
  if not url:
    print(
      f'No url found for {sub}, visit https://run-rx.com/pricing to sign up.')
    return
  print(f'\nOpening your browser. Or, you can manually visit:\n\n{url}')
  browser.open_browser(url)

  print("""
Thank you! Your subscription will be activated as soon as your payment is
processed (that may take a few minutes).""")


def success():
  print("""

Subscription activated! You can now continue using rx.
""")



import argparse
import datetime
import sys
from typing import List

from google.protobuf import json_format
import yaml

from rx.client import grpc_helper
from rx.client import trex_client
from rx.client.commands import command
from rx.client.configuration import config_base
from rx.proto import rx_pb2


class WorkspaceInfoCommand(command.Command):
  """Get info about the current workspace."""

  def run(self) -> int:
    try:
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = trex_client.create_authed_client(ch, self.local_config)
        info = client.get_info(self.remote_config.workspace_id)
    except config_base.ConfigNotFoundError as e:
      print('No workspace found.')
      return -1
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code

    target_env = json_format.MessageToDict(info.environment)

    # TODO: check if info matches the remote config.
    print(f"""Workspace info for {self.local_config.cwd}

Remote config:\n\n{yaml.safe_dump(target_env)}
Commands run this month:
""")
    for cmd in info.history:
      start = datetime.datetime.fromtimestamp(cmd.start_ts)
      print(f'  {start}\t{cmd.cmd}')
    return 0


def _run_cmd(args: List[str]) -> int:
  del args
  cmd = WorkspaceInfoCommand()
  return cmd.run()


def add_parser(subparsers: argparse._SubParsersAction):
  (
    subparsers
    .add_parser('workspace-info', help='Gets info about the current workspace')
    .set_defaults(func=_run_cmd)
  )


if __name__ == '__main__':
  print('Call exec.')



"""rx usage

To set up this directory on a remote host:

    rx init

You only have to run this once per directory, similar to `git init`.

To run a command on the remote host:

    rx CMD

Run 'rx --help' for more options or visit https://docs.run-rx.com.
"""
import argparse
import sys
import tempfile
from typing import List

from absl import app
from absl import flags
from absl import logging
from absl.flags import argparse_flags

from rx.client import worker_client
from rx.client.commands import init
from rx.client.commands import runner
from rx.client.commands import stop
from rx.client.commands import subscribe
from rx.client.commands import workspace_info
from rx.client.configuration import local


def _get_version(args: List[str]) -> int:
  del args
  print(local.VERSION)
  return 0


def get_subcommand_parser(
    parser: argparse_flags.ArgumentParser) -> argparse._SubParsersAction:
  subparsers = parser.add_subparsers()
  (
    subparsers
    .add_parser('help', help='Show help message for a given command')
    .set_defaults(func=lambda _: parser.print_help())
  )
  init.add_parser(subparsers)
  runner.add_parser(subparsers)
  stop.add_parser(subparsers)
  subscribe.add_parsers(subparsers)
  workspace_info.add_parser(subparsers)
  (
    subparsers
    .add_parser('version', help='Gets the version of the rx client')
    .set_defaults(func=_get_version)
  )
  return subparsers


def main(argv):
  logging.get_absl_handler().python_handler.use_absl_log_file(
    program_name='rx', log_dir=tempfile.gettempdir())

  parser = argparse_flags.ArgumentParser(
    description=(
      'rx is a cli interface for seamless hybrid development. Develop locally, '
      'then run locally or in the cloud.'))
  subparsers = get_subcommand_parser(parser)
  if len(argv) == 1:
    parser.print_help()
    return -1

  if argv[1] in subparsers.choices:
    ns, remainder = parser.parse_known_args(argv[1:])
    try:
      return ns.func(remainder)
    except KeyboardInterrupt:
      return worker_client.SIGINT_CODE
  else:
    # argparse doesn't like it when it doesn't recognize anything.
    return runner.RunCommand(argv[1:]).run()


def parse_flags_with_usage(args):
  try:
    return flags.FLAGS(args)
  except flags.Error as error:
    cmd = ' '.join(args[1:])
    print(
      f'{error}\n\nIf you are attempting to run a command on a remote machine, '
      f'try using quotes:\n\n\trx \'{cmd}\'')
    sys.exit(1)


# Run from rx.__main__.
def run():
  # Override absl's flag parser method with one that doesn't have a hard-coded
  # error message and then call sys.exit.
  app.run(main, flags_parser=parse_flags_with_usage)



import argparse
import pathlib
import sys
from typing import List

from absl import flags
from absl import logging

from rx.client import grpc_helper
from rx.client import login
from rx.client import menu
from rx.client import trex_client
from rx.client.commands import command
from rx.client.configuration import config_base
from rx.client.configuration import local

_DRY_RUN = flags.DEFINE_bool(
  'dry-run', False, 'Shows a list of files that would be uploaded by rx init')


class InitCommand(command.Command):
  """Initialize (or reinitialize) the remote."""

  def __init__(self):
    super().__init__()
    if self._rxroot:
      self._config_exists = local.get_local_config_path(self._rxroot).exists()
      self._user_info_exists = not login.needs_login(self._rxroot)
    else:
      # Fall back on using this dir for rxroot.
      self._rxroot = pathlib.Path.cwd()
      self._config_exists = False
      self._user_info_exists = False

  @property
  def rxroot(self) -> pathlib.Path:
    return self._rxroot

  def _show_init_message(self):
    underline = '=' * len(str(self._rxroot))
    print(f"""
rx setup for {self._rxroot}
============={underline}
"""
)
    steps = []
    if self._config_exists:
      reinit = menu.bool_prompt(
        'Looks like you already have an rx workspace. Would you like to stop '
        'that one\nand start a new one?', 'y')
      if not reinit:
        print('Okay, goodbye!')
        sys.exit(0)
      message = 'Got it. To re-init this workspace, rx will:\n\n'
      steps.append('Shut down your existing virtual machine.')
    else:
      message = 'To set up rx, this command will:\n\n'
    if not self._user_info_exists:
      steps.append('Create your rx account (or log in).')
    steps.append('Set up a virtual machine on AWS.')
    steps.append(f'Copy the files in {self._rxroot} to your virtual machine.')
    for num, step in enumerate(steps, 1):
      message += f'{num}. {step}\n'
    message += ('\nWould you like to continue?')
    if not menu.bool_prompt(message, 'y'):
      print('Okay, goodbye!')
      sys.exit(0)

  def _set_up_user(
      self, client: trex_client.Client) -> bool:
    """Returns whether to continue."""
    if not self._user_info_exists:
      ready_to_login = menu.bool_prompt(
        f"""
First rx will need to open a browser window for you to log in. rx uses Google's
oauth to associate your email with your rx account.

Press y to continue:""", 'y')
      if not ready_to_login:
        print('Okay, goodbye!')
        return False
    client.create_user_or_log_in()
    return True

  def run(self) -> int:
    self._show_init_message()
    if self._config_exists:
      logging.info('Workspace already exists, resetting it.')
    config = local.create_local_config(self._rxroot)
    client = None
    try:
      with grpc_helper.get_channel(config_base.TREX_HOST.value) as ch:
        client = trex_client.Client(ch, config, auth_metadata=None)
        if _DRY_RUN.value:
          client.dry_run()
          return 0

        cont = self._set_up_user(client)
        if not cont:
          return 0

        ready_to_upload = menu.bool_prompt(
          f"""
Upload your code
================

First rx is going to start a virtual machine in the cloud for your user. This is
your private machine and its state will be saved between commands.

The source code in this directory will be copied to your virtual machine. To
check what will be uploaded, you can rerun this command with --dry-run and
modify .rxignore to exclude anything you don't want copied.

Are you sure you want to upload {self._rxroot} to the cloud?""", 'y')
        if not ready_to_upload:
          print('Okay, goodbye!')
          return 0
        if not menu._QUIET.value:
          print('Great! Let\'s get down to business.')
        return client.init()
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code

  def second_try(self, client: trex_client.Client) -> int:
    try:
      return client.init()
    except trex_client.TrexError as e:
      sys.stderr.write(f'{e}\n')
      sys.stderr.flush()
      return e.code


def _run_cmd(args: List[str]) -> int:
  del args
  cmd = InitCommand()
  return cmd.run()


def add_parser(subparsers: argparse._SubParsersAction):
  (
    subparsers
    .add_parser('init', help='Allocates and sets up a new workspace in AWS')
    .set_defaults(func=_run_cmd)
  )


if __name__ == '__main__':
  print('Call exec.')
